
/* -------------------------------------------------------------------------- */
/*
        programname: nph-count
        description: Graphical counter on WWW pages. Should be used as
             <IMG SRC="/cgi-bin/nph-count?width=5&link=/dummy/index.html">

        author: H. Withagen; www@eeb.ele.tue.nl
        creation : 05-05-1995
        last modification : 13-06-1995

*/   
/* -------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include "util.h"
#include "gd/gd.h"


#ifndef TRUE
#define TRUE 		1
#endif

#ifndef FALSE
#define FALSE 		0
#endif

#ifndef RAND_MAX
#define RAND_MAX	2147483646
#endif

#ifndef SIGALRM
#define SIGALRM		14
#endif

#define MAX_LENGTH	256
#define DIGIT_WIDTH	15
#define DIGIT_HEIGHT	20
#define ALARM_INTERVAL	5


typedef struct {
    char *name;
    char *val;
} entry;


/* local functions */
int retrieve_arguments(entry *entries);
int link_search(char *link);
int make_gif(int width, char *count_string);
gdImagePtr load_digit(char digit_char);
int random_number(int width);
static void sig_alarm(int signo);
void get_lock(void);
void release_lock(void);
void exit_with_http_error(char *message);

/* --------------------------------------------------------------------------- */

int main(int argc, char *argv[]) 
{
  int width, arg_count, link_count;
  char str[256];
  entry entries[32];


  /* program the alarm clock to prevent programs
     from hanging themself up */
  signal(SIGALRM, sig_alarm);
  alarm(ALARM_INTERVAL);

  /* get lock for the running version of the program */
  get_lock();

  /* open logging facility */
  openlog(argv[0], LOG_PID, LOG_MAIL);


  arg_count = retrieve_arguments(entries);

#if 0
  /* still here for testing purposes */
  printf("Content-type: text/html%c%c<pre>",10,10);
  for (width = 0; width < arg_count; width++)
  {
    printf("%s = %s\n", entries[width].name, entries[width].val);
    if (!strlen(entries[width].val))
      printf("Value argument is empty (stringlength is 0)\n");
  }
  printf("</pre>");
  return 0;
#endif

  switch (arg_count) {
    case 1: width = 0;
            if ( !strcmp(entries[0].name, "link") && strlen(entries[0].val) )
              if ( !strcmp(entries[0].val, "random") )
                link_count = random_number(width);
              else
                link_count = link_search(entries[0].val);
            else
              exit_with_http_error("Bad arguments");

            break;
    case 2: if ( !strcmp(entries[0].name, "width") && strlen(entries[0].val) )
              width = atoi(entries[0].val);
            else
              exit_with_http_error("Bad arguments");

            if ( !strcmp(entries[1].name, "link") && strlen(entries[1].val) )
              if ( !strcmp(entries[1].val, "random") )
                link_count = random_number(width);
              else
                link_count = link_search(entries[1].val);
            else
              exit_with_http_error("Bad arguments");
            
            break;
  }

  if (!link_count)
    exit_with_http_error("Internal error image counter");

  sprintf(str, "%d", link_count);  
  if (make_gif(width, str))
    exit_with_http_error("Internal error image counter");
  
  closelog();
  release_lock();
  return 0;
}

/* --------------------------------------------------------------------------- */

/* retrieve arguments from both FORM (POST) and 
   GET inputs. For the GET input it should be of 
   the form /cgi-bin/progname?name1=val1&name2=val2 */
int retrieve_arguments(entry *entries)
{
  register int x,m=0;
  int cl;
  char line[256], *line_ptr, str[256];

  if (!strcmp((char *)getenv("REQUEST_METHOD"),"POST"))
  {
    cl = atoi((char *)getenv("CONTENT_LENGTH"));

    for(x=0;cl && (!feof(stdin));x++) 
    {
      m=x;
      entries[x].val = fmakeword(stdin,'&',&cl);
      plustospace(entries[x].val);
      unescape_url(entries[x].val);
      entries[x].name = makeword(entries[x].val,'=');
    }
    return m+1;
  }
  else
  {
    strcpy(line, (char*)getenv("QUERY_STRING"));    
    line_ptr = line;
    m = 0;
    x = 0;
    while (*line_ptr)
      if ((*line_ptr) == '&')
      {
        str[x] = '\0';
        entries[m].val = (char *)malloc(sizeof(char) * (strlen(line) + 1));
        strcpy(entries[m++].val, str);
        x = 0;
        line_ptr++;
      }
      else
      {
        str[x++] = *line_ptr;
        line_ptr++;
      }

    str[x] = '\0';
    entries[m].val = (char *)malloc(sizeof(char) * (strlen(line) + 1));
    strcpy(entries[m].val, str);

    for (x=0; x<=m; x++)
    {
      plustospace(entries[x].val);
      unescape_url(entries[x].val);
      entries[x].name = makeword(entries[x].val,'=');
    }
  
    return m+1;
  }

}

/* --------------------------------------------------------------------------- */

int link_search(char *link)
{
  char file_name[MAX_LENGTH], line[MAX_LENGTH];
  char **link_array, **link_array_ptr;
  int count, found = FALSE, number_of_lines;
  int found_count, *count_array, *count_array_ptr;
  FILE *fp;

  /* make sure we are not interrupted in this function */
  alarm(0);

  sprintf(file_name,"%s", COUNT_FILE);
  if ( !(fp = fopen(file_name, "r")) )
  {
    /* file could not be opened or doesn't exist 
       we'll assume, it doesn't exist and create a new one */
       if ( !(fp = fopen(file_name, "w")) )
         return 0;
       fprintf(fp, "1\n%s\t1\n", link);
       fclose(fp);
       return 1;
  }

  /* read the number of entries in the file */
  fgets(line, MAX_LENGTH, fp);
  sscanf(line, "%d", &number_of_lines);

  link_array = (char **)malloc( number_of_lines * sizeof(char *) );
  link_array_ptr = link_array;
  if (!link_array)
    return 0;
  count_array = (int *)malloc( number_of_lines * sizeof(int) );
  count_array_ptr = count_array;
  if (!count_array)
    return 0;

  for (count = 0; count < number_of_lines; count++, link_array_ptr++, count_array_ptr++)
  {
    /* allocate space for a new entry */
    *link_array_ptr = (char *)malloc( MAX_LENGTH * sizeof(char) );
    if (!*link_array_ptr)
      return 0;

    /* read the next line from file */
    fgets(line, MAX_LENGTH, fp);

    /* copy the information into the arrays */
    sscanf(line, "%s %d", *link_array_ptr, count_array_ptr);

    /* check if this is the link we are looking for */
    if ( !strcmp(*link_array_ptr, link) )
    {
      found = TRUE;
      (*count_array_ptr)++;
      found_count = *count_array_ptr;
    }
  }

  fclose(fp);


  /* write out file again */
  if ( !(fp = fopen(file_name, "w")) )
    return 0;

  /* check if a new line should be added */
  if (found)
    fprintf(fp,"%d\n", number_of_lines);
  else
    fprintf(fp,"%d\n", number_of_lines+1);

  link_array_ptr = link_array;
  count_array_ptr = count_array;
  for (count = 0; count < number_of_lines; count++, link_array_ptr++, count_array_ptr++)
  {
    fprintf(fp, "%s\t%d\n", *link_array_ptr, *count_array_ptr);
    free(*link_array_ptr);
  }

  if (!found)
    fprintf(fp, "%s\t%d\n", link, 1);

  free(link_array);
  free(count_array);
  fclose(fp);

  alarm(ALARM_INTERVAL);
   
  if (found)
    return found_count;
  else
    return 1;
}

/* --------------------------------------------------------------------------- */

int make_gif(int width, char *count_string)
{
  gdImagePtr image_ptr, digit_ptr;
  char *str_ptr = count_string;
  int count, digit_count = 0;


  /* adjust image width */
  if ( width > 10 )
    width = 10;
  if ( strlen(count_string) > width )
    width = strlen(count_string);


  /* create new image to contain the new counter image */
  if ( !(image_ptr = gdImageCreate(width * DIGIT_WIDTH, DIGIT_HEIGHT)) )
  {
    syslog(LOG_DEBUG, "unable to create new counter image");
    return 1;
  }

  if ( !(digit_ptr = load_digit('0')) )
    return 1;
 
  for (count = 0; count < (width - strlen(count_string)); count++)
  {
    gdImageCopy(image_ptr, digit_ptr, digit_count * DIGIT_WIDTH, 0, 0, 0, DIGIT_WIDTH, DIGIT_HEIGHT);
    digit_count++;
  }  

  for (count = 0; count < strlen(count_string); count++, str_ptr++)
  {
    if ( !(digit_ptr = load_digit(*str_ptr)) )
      return 1;

    gdImageCopy(image_ptr, digit_ptr, digit_count * DIGIT_WIDTH, 0, 0, 0, DIGIT_WIDTH, DIGIT_HEIGHT);
    digit_count++;
  }

  fputs("HTTP/1.0 200 OK\n", stdout);
  fputs("Server: Netsite-Communications/1.0\n", stdout);
  fputs("Content-type: image/gif\n\n", stdout);
  gdImageGif(image_ptr, stdout);

  return 0;
}

/* --------------------------------------------------------------------------- */

gdImagePtr load_digit(char digit_char)
{
  char file_name[256], str[256];
  FILE *fp;
  gdImagePtr digit_ptr;

  sprintf(file_name,"%s%c.gd", PIC_PATH, digit_char);
  if ( !(fp = fopen(file_name, "r")) )
  {
    sprintf(str, "could not open digit file %c.gd", digit_char);
    syslog(LOG_DEBUG, "%s", str);
    return NULL;
  }

  if ( !(digit_ptr = gdImageCreateFromGd(fp)) )
  {
    sprintf(str, "%c.gd digit file doesn't contain a gd-image", digit_char);
    syslog(LOG_DEBUG, "%s", str);  
    return NULL;
  }

  fclose(fp);
  return digit_ptr;
}

/* --------------------------------------------------------------------------- */

/* Function returns a random number with the specified width.
   If width = 0, an number with arbitrary width will be returned */

int random_number(int width)
{
  srand(time(NULL));
  if (width)
    return (( (double)rand()/(double)RAND_MAX ) * pow( 10.0, (double)width )); 
  else
    return rand();
}

/* --------------------------------------------------------------------------- */

/* alarm function is only reached if something went wrong */
   
static void sig_alarm(int signo)
{
  syslog(LOG_DEBUG, "nph-count alarm activated, releasing lock; exiting");  
  release_lock();
  exit(0);
}

/* --------------------------------------------------------------------------- */

void get_lock(void)
{
  int fd;
  
  /* try creating a lock file */
  while ( (fd = open("/tmp/nph-count.lock", O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR)) < 0 );
  close(fd);
}

/* --------------------------------------------------------------------------- */

void release_lock(void)
{
  unlink("/tmp/nph-count.lock");
}

/* --------------------------------------------------------------------------- */

void exit_with_http_error(char *message)
{
  char str[64];

  sprintf(str, "HTTP/1.0 500 %s\n", message);
  fputs(str, stdout);
  release_lock();

  exit(1);
}

/* --------------------------------------------------------------------------- */

